from __future__ import annotations

import fcntl
import io
import logging
import struct
from pathlib import Path
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from collections.abc import Generator
from contextlib import suppress

from labels.pkg.cataloger.redhat.rpmdb.rpmdb_interface import RpmDBInterface
from labels.utils.exceptions import InvalidDBFormatError

# Constants
NDB_SLOT_ENTRIES_PER_PAGE = 4096 // 16  # 16 bytes per ndbSlotEntry
NDB_HEADER_MAGIC = ord("R") | (ord("p") << 8) | (ord("m") << 16) | (ord("P") << 24)
NDB_DB_VERSION = 0
LOGGER = logging.getLogger(__name__)


def syscall_flock(fd: int, how: int) -> None:
    fcntl.flock(fd, how)


# Struct formats
NDB_HEADER_STRUCT = struct.Struct("<8I")  # 8 unsigned 32-bit integers
NDB_SLOT_ENTRY_STRUCT = struct.Struct("<4I")  # 4 unsigned 32-bit integers
NDB_BLOB_HEADER_STRUCT = struct.Struct("<4I")  # 4 unsigned 32-bit integers


class RpmNDB(RpmDBInterface):
    def __init__(self, file: io.BufferedReader, slots: list[dict[str, int]]) -> None:
        self.file = file
        self.slots = slots

    @classmethod
    def open(cls, path: str) -> RpmNDB:
        file = Path(path).open("rb")  # noqa: SIM115
        try:
            syscall_flock(file.fileno(), fcntl.LOCK_SH)
        except Exception as exc:
            file.close()
            raise exc  # noqa: TRY201

        # Read NDB header
        hdr_data = file.read(NDB_HEADER_STRUCT.size)
        if len(hdr_data) != NDB_HEADER_STRUCT.size:
            file.close()
            error_msg = "Failed to read metadata"
            raise InvalidDBFormatError(error_msg)

        hdr_fields = NDB_HEADER_STRUCT.unpack(hdr_data)
        hdr = {
            "HeaderMagic": hdr_fields[0],
            "NDBVersion": hdr_fields[1],
            "NDBGeneration": hdr_fields[2],
            "SlotNPages": hdr_fields[3],
            # Remaining fields are ignored
        }

        if (
            hdr["HeaderMagic"] != NDB_HEADER_MAGIC
            or hdr["SlotNPages"] == 0
            or hdr["NDBVersion"] != NDB_DB_VERSION
        ):
            file.close()
            error_msg = "Invalid or unsupported NDB format"
            raise InvalidDBFormatError(error_msg)

        if hdr["SlotNPages"] > 2048:
            file.close()
            error_msg = f"Slot page limit exceeded: {hdr['SlotNPages']:x}"
            raise InvalidDBFormatError(error_msg)

        # Read slot entries
        num_slots = hdr["SlotNPages"] * NDB_SLOT_ENTRIES_PER_PAGE - 2  # First two slots are headers
        slots_data = file.read(num_slots * NDB_SLOT_ENTRY_STRUCT.size)
        if len(slots_data) != num_slots * NDB_SLOT_ENTRY_STRUCT.size:
            file.close()
            error_msg = "Failed to read NDB slot pages"
            raise InvalidDBFormatError(error_msg)

        slots = []
        for i in range(num_slots):
            slot_data = slots_data[
                i * NDB_SLOT_ENTRY_STRUCT.size : (i + 1) * NDB_SLOT_ENTRY_STRUCT.size
            ]
            slot_fields = NDB_SLOT_ENTRY_STRUCT.unpack(slot_data)
            slot = {
                "SlotMagic": slot_fields[0],
                "PkgIndex": slot_fields[1],
                "BlkOffset": slot_fields[2],
                "BlkCount": slot_fields[3],
            }
            slots.append(slot)

        return cls(file, slots)

    def close(self) -> None:
        with suppress(Exception):
            syscall_flock(self.file.fileno(), fcntl.LOCK_UN)
        self.file.close()

    def read(self) -> Generator[bytes, None, None]:
        ndb_slot_magic = ord("S") | (ord("l") << 8) | (ord("o") << 16) | (ord("t") << 24)
        ndb_blob_magic = ord("B") | (ord("l") << 8) | (ord("b") << 16) | (ord("S") << 24)

        for slot in self.slots:
            if slot["SlotMagic"] != ndb_slot_magic:
                LOGGER.error("Bad slot magic %s", slot["SlotMagic"])
                return
            # Empty slot?
            if slot["PkgIndex"] == 0:
                continue
            # Seek to Blob
            offset = slot["BlkOffset"] * NDB_BLOB_HEADER_STRUCT.size
            try:
                self.file.seek(offset, io.SEEK_SET)
            except OSError:
                LOGGER.exception("Failed to seek to blob")
                return

            # Read Blob Header
            blob_header_data = self.file.read(NDB_BLOB_HEADER_STRUCT.size)
            if len(blob_header_data) != NDB_BLOB_HEADER_STRUCT.size:
                LOGGER.error("Failed to read blob header")
                return
            blob_header_fields = NDB_BLOB_HEADER_STRUCT.unpack(blob_header_data)
            blob_header = {
                "BlobMagic": blob_header_fields[0],
                "PkgIndex": blob_header_fields[1],
                "BlobCkSum": blob_header_fields[2],
                "BlobLen": blob_header_fields[3],
            }
            if blob_header["BlobMagic"] != ndb_blob_magic:
                LOGGER.error(
                    "Unexpected NDB blob Magic for pkg %s: %s",
                    slot["PkgIndex"],
                    blob_header["BlobMagic"],
                )
                return
            if blob_header["PkgIndex"] != slot["PkgIndex"]:
                LOGGER.error("Failed to find NDB blob for pkg %s", slot["PkgIndex"])
                return

            # Read Blob Content
            blob_len = blob_header["BlobLen"]
            blob_entry = self.file.read(blob_len)
            if len(blob_entry) != blob_len:
                LOGGER.error("Failed to read blob content")
                return
            yield blob_entry



from labels.artifact.relationship import (
    Relationship,
)
from labels.file.resolver import Resolver
from labels.format.common import (
    process_packages,
)
from labels.format.cyclone_dx.file_builder import (
    format_cyclonedx_sbom,
)
from labels.format.fluid import (
    format_fluid_sbom,
)
from labels.format.spdx.file_builder import (
    format_spdx_sbom,
)
from labels.model.core import (
    Package,
    SbomConfig,
)


def format_sbom(
    *,
    packages: list[Package],
    relationships: list[Relationship],
    config: SbomConfig,
    resolver: Resolver,
) -> None:
    packages = process_packages(packages)
    match config.output_format:
        case "fluid-json":
            format_fluid_sbom(
                packages=packages,
                relationships=relationships,
                config=config,
                resolver=resolver,
            )
        case "cyclonedx-json":
            format_cyclonedx_sbom(
                packages=packages,
                relationships=relationships,
                config=config,
                resolver=resolver,
            )
        case "cyclonedx-xml":
            format_cyclonedx_sbom(
                packages=packages,
                relationships=relationships,
                config=config,
                resolver=resolver,
            )
        case "spdx-json":
            format_spdx_sbom(
                packages=packages,
                _relationships=relationships,
                file_format="json",
                config=config,
                resolver=resolver,
            )
        case "spdx-xml":
            format_spdx_sbom(
                packages=packages,
                _relationships=relationships,
                file_format="xml",
                config=config,
                resolver=resolver,
            )



import logging
import textwrap
from typing import Unpack

import click

from labels.core.configurator import (
    ScanArgs,
    build_labels_config_from_args,
    build_labels_config_from_file,
)
from labels.core.scanner import execute_labels_scan

LOGGER = logging.getLogger(__name__)


def show_fluid_attacks_banner() -> None:
    logo = textwrap.dedent(
        """
         ───── ⌝
        |    ⌝|  Fluid Attacks
        |  ⌝  |  We hack your software.
         ─────
        """,
    )
    click.secho(logo, fg="red")


@click.group(help="Fluid labels CLI")
def cli() -> None:
    """Entry point for the command-line interface (CLI) of the application.

    This function serves as the main entry point for the CLI, grouping
    all the available commands under a single interface.

    Available commands:
    - config: Load and process a configuration file.
    - scan: Perform a scan based on the provided arguments and options.
    """


@click.command(
    name="config",
    short_help="Load and process a YAML configuration file.",
    epilog="Check out our docs at https://help.fluidattacks.com/portal/en/kb",
)
@click.argument(
    "config_path",
    type=click.Path(exists=True, readable=True, path_type=str),
    required=True,
    metavar="config_path",
)
def run_labels_from_config(config_path: str) -> None:
    """Execute SBOM scan from a configuration file.

    \b
    ARGUMENT:
        config_path:  Path to the YAML configuration file (.yaml or .yml).

    \b
    USAGE EXAMPLES:
        - labels config ./config.yaml   # Load and execute the SBOM scan using config.yaml
        - labels config ./config.yml    # Load and execute the SBOM scan using config.yml

    The configuration file must be in YAML format (.yaml or .yml).
    """  # noqa: D301
    show_fluid_attacks_banner()
    sbom_config = build_labels_config_from_file(config_path)
    execute_labels_scan(sbom_config)


@click.command(
    "scan",
    short_help="Perform an SBOM scan using the specified args.",
    epilog="Check out our docs at https://help.fluidattacks.com/portal/en/kb",
)
@click.option(
    "--source",
    type=click.Choice(["docker", "dir", "docker-daemon", "ecr"], case_sensitive=False),
    required=True,
    default="dir",
    show_default=True,
    help="Specify the source of the scan: 'docker', 'dir', 'docker-daemon', or 'ecr'.",
)
@click.option(
    "--output",
    default="my_sbom",
    type=click.STRING,
    show_default=True,
    help="Output filename for the scan results.",
)
@click.option(
    "--format",
    type=click.Choice(
        ["fluid-json", "cyclonedx-json", "spdx-json", "cyclonedx-xml", "spdx-xml"],
        case_sensitive=False,
    ),
    default="fluid-json",
    help="Output format for the resulting file.",
)
@click.option(
    "--docker-user",
    default=None,
    type=click.STRING,
    help="Docker registry username, if required.",
)
@click.option(
    "--docker-password",
    default=None,
    type=click.STRING,
    help="Docker registry password, if required.",
)
@click.option(
    "--aws-external-id",
    default=None,
    type=click.STRING,
    help="AWS external ID for ECR access, if necessary.",
)
@click.option(
    "--aws-role",
    default=None,
    type=click.STRING,
    help="AWS role ARN for ECR access, if necessary.",
)
@click.option(
    "--debug",
    is_flag=True,
    default=False,
    show_default=True,
    help="Enable debug mode for detailed logging.",
)
@click.argument(
    "target",
    type=click.Path(exists=True, readable=True, path_type=str),
    required=True,
)
def run_labels_from_args(target: str, **kwargs: Unpack[ScanArgs]) -> None:
    """Execute SBOM scan with arguments.

    \b
    ARGUMENT:
        target:  The target to scan, depending on the selected source:

    \b
    USAGE EXAMPLES:
        - labels scan my-image:latest --from docker # Scan a Docker image
        - labels scan ./my-project --from dir # Scan a local directory
        - labels scan my-image:latest --from docker-daemon # Scan a running local image
        - labels scan 1234567890.dkr.ecr.us-east-1.amazonaws.com/my-image --from ecr\
 # Scan an ECR image
    """  # noqa: D301
    show_fluid_attacks_banner()
    sbom_config = build_labels_config_from_args(target, **kwargs)
    execute_labels_scan(sbom_config)


cli.add_command(run_labels_from_config)
cli.add_command(run_labels_from_args)

if __name__ == "__main__":
    cli(prog_name="sbom")



from datetime import UTC, datetime

from packageurl import (
    PackageURL,
)

from labels.advisories import (
    roots as advisories,
)
from labels.internal.package_information.java import (
    MavenPackageInfo,
    get_maven_package_info,
    search_maven_package,
)
from labels.model.core import (
    Artifact,
    Digest,
    HealthMetadata,
    Package,
)
from labels.pkg.cataloger.common import (
    infer_algorithm,
)
from labels.pkg.cataloger.java.model import (
    JavaArchive,
    JavaPomProject,
    JavaPomProperties,
)
from labels.utils.licenses.validation import (
    validate_licenses,
)


def _update_advisories(package: Package) -> None:
    pkg_advisories = advisories.get_vulnerabilities("maven", package.name, package.version)
    if pkg_advisories:
        package.advisories = pkg_advisories


def looks_like_group_id(group_id: str) -> bool:
    return "." in group_id


def remove_osgi_directives(group_id: str) -> str:
    return group_id.split(";")[0]


def clean_group_id(group_id: str) -> str:
    return remove_osgi_directives(group_id).strip()


def group_id_from_pom_properties(properties: JavaPomProperties | None) -> str:
    if not properties:
        return ""
    if properties.group_id:
        return clean_group_id(properties.group_id)
    if properties.artifact_id and looks_like_group_id(properties.artifact_id):
        return clean_group_id(properties.artifact_id)
    return ""


def group_id_pom_project(project: JavaPomProject | None) -> str:
    if not project:
        return ""
    if project.group_id:
        return clean_group_id(project.group_id)
    if project.artifact_id and looks_like_group_id(project.artifact_id):
        return clean_group_id(project.artifact_id)
    if project.parent:
        if project.parent.group_id:
            return clean_group_id(project.parent.group_id)
        if looks_like_group_id(project.parent.artifact_id):
            return clean_group_id(project.parent.artifact_id)
    return ""


def group_id_from_java_metadata(_pkg_name: str, metadata: JavaArchive | None) -> str | None:
    if metadata:
        if group_id := group_id_from_pom_properties(metadata.pom_properties):
            return group_id
        if group_id := group_id_pom_project(metadata.pom_project):
            return group_id
    return None


def package_url(name: str, version: str, metadata: JavaArchive | None = None) -> str:
    group_id = name
    if (g_id := group_id_from_java_metadata(name, metadata)) and g_id:
        group_id = g_id
    return PackageURL(
        type="maven",
        namespace=group_id,
        name=name,
        version=version,
        qualifiers=None,
        subpath="",
    ).to_string()


def _get_health_metadata_artifact(
    current_package: MavenPackageInfo | None,
) -> Artifact | None:
    if current_package and current_package.jar_url:
        digest_value = current_package.hash or None
        return Artifact(
            url=current_package.jar_url,
            integrity=Digest(
                algorithm=infer_algorithm(digest_value),
                value=digest_value,
            ),
        )
    return None


def _get_group_id(package: Package) -> str | None:
    if isinstance(package.metadata, JavaArchive) and (
        g_id := group_id_from_java_metadata(package.name, package.metadata)
    ):
        return g_id
    if package_candidate := search_maven_package(package.name, package.version):
        return package_candidate.group
    return None


def _set_health_metadata(
    package: Package,
    maven_package: MavenPackageInfo,
    current_package: MavenPackageInfo | None,
) -> None:
    authors = maven_package.authors or []
    package.health_metadata = HealthMetadata(
        latest_version=maven_package.latest_version,
        latest_version_created_at=datetime.fromtimestamp(maven_package.release_date, tz=UTC)
        if maven_package.release_date
        else None,
        authors=", ".join(authors) if authors else None,
        artifact=_get_health_metadata_artifact(current_package),
    )


def complete_package(package: Package) -> Package:
    _update_advisories(package)
    group_id = _get_group_id(package)
    if not group_id:
        return package

    maven_package = get_maven_package_info(group_id, package.name)
    if not maven_package:
        if package_candidate := search_maven_package(package.name, package.version):
            maven_package = get_maven_package_info(package_candidate.group, package.name)
        if not maven_package:
            return package

    current_package = get_maven_package_info(group_id, package.name, package.version)

    _set_health_metadata(package, maven_package, current_package)

    maven_package_licenses = maven_package.licenses or []
    if maven_package_licenses:
        package.licenses = validate_licenses(
            [package_license for package_license in maven_package_licenses if package_license],
        )
    return package



import logging

import tree_sitter_json
from tree_sitter import (
    Language as TLanguage,
)
from tree_sitter import (
    Node,
    Parser,
)

from labels.internal.collection.types import (
    IndexedDict,
    IndexedList,
    ParsedValue,
)
from labels.utils.exceptions import (
    UnexpectedNodeError,
)

LOGGER = logging.getLogger(__name__)


def _handle_array_node(node: Node) -> tuple[Node, IndexedList[ParsedValue]]:
    data: IndexedList[ParsedValue] = IndexedList(node)
    for child in node.children:
        if child.type not in ("[", "]", ","):
            try:
                value_node, value = handle_json_node(child)
                data.append((value, value_node))
            except UnexpectedNodeError:
                LOGGER.exception(
                    "Unexpected node type encountered while handling array node: %s",
                    child.type,
                    extra={
                        "extra": {
                            "node_type": child.type,
                        },
                    },
                )
                continue
    return node, data


def _handle_object_node(node: Node) -> tuple[Node, IndexedDict[str, ParsedValue]]:
    data: IndexedDict[str, ParsedValue] = IndexedDict(node)
    for child in node.children:
        if child.type == "pair":
            key_n, _, value_n = child.children
            if not key_n.text:
                continue
            key = key_n.text[1:-1].decode("utf-8")
            try:
                value_node, value_value = handle_json_node(value_n)
                data[(key, key_n)] = (value_value, value_node)
            except UnexpectedNodeError:
                LOGGER.exception(
                    "Unexpected node type encountered while handling object node: %s",
                    child.type,
                    extra={
                        "extra": {
                            "node_type": child.type,
                        },
                    },
                )
                continue
    return node, data


def handle_json_node(node: Node) -> tuple[Node, ParsedValue]:
    value: tuple[Node, ParsedValue]
    match node.type:
        case "array":
            value = _handle_array_node(node)
        case "object":
            value = _handle_object_node(node)
        case "string":
            node_value = node.text[1:-1].decode("utf-8") if node.text else ""
            value = node, node_value
        case "number":
            node_value = node.text.decode("utf-8") if node.text else "0"
            try:
                value = node, int(node_value)
            except ValueError:
                value = node, float(node_value)
        case "true":
            value = node, True
        case "false":
            value = node, False
        case "null":
            value = node, None
        case _:
            raise UnexpectedNodeError(node)
    return value


def parse_json_with_tree_sitter(
    json: str,
) -> IndexedDict[str, ParsedValue] | IndexedList[ParsedValue]:
    parser_language = TLanguage(tree_sitter_json.language())
    parser = Parser(parser_language)
    result = parser.parse(json.encode("utf-8"))
    value: ParsedValue = IndexedDict()
    for child in result.root_node.children:
        try:
            _, value = handle_json_node(child)
        except UnexpectedNodeError:
            LOGGER.exception(
                "Unexpected node type encountered: %s",
                child.type,
                extra={
                    "extra": {
                        "node_type": child.type,
                        "json": json,
                    },
                },
            )
            continue
    if value is None or not isinstance(value, IndexedDict | IndexedList):
        LOGGER.warning(
            "JSON parsing failed.",
            extra={"extra": {"json": json}},
        )
        return IndexedDict()
    return value



from packageurl import (
    PackageURL,
)

from labels.advisories import (
    roots as advisories,
)
from labels.internal.package_information.rust import (
    CRATES_ENDPOINT,
    CargoPackage,
    Version,
    get_cargo_package,
)
from labels.model.core import (
    Artifact,
    Digest,
    HealthMetadata,
    Package,
)
from labels.pkg.cataloger.common import (
    infer_algorithm,
)
from labels.utils.licenses.validation import (
    validate_licenses,
)


def package_url(name: str, version: str) -> str:
    return PackageURL(  # type: ignore
        type="cargo",
        namespace="",
        name=name,
        version=version,
        qualifiers=None,
        subpath="",
    ).to_string()


def _get_artifact(
    current_package: Version | None,
) -> Artifact | None:
    if current_package:
        digest_value = current_package.get("checksum")
        return Artifact(
            url=f"{CRATES_ENDPOINT}{current_package['dl_path']}",
            integrity=Digest(
                algorithm=infer_algorithm(digest_value),
                value=digest_value,
            ),
        )
    return None


def _set_health_metadata(
    package: Package,
    current_package: Version | None,
    cargo_package: CargoPackage,
) -> None:
    crate_info = cargo_package.get("crate", {})
    max_stable_version = crate_info.get("max_stable_version")
    updated_at = crate_info.get("updated_at")
    published_by = cargo_package["versions"][0].get("published_by")

    package.health_metadata = HealthMetadata(
        latest_version=max_stable_version,
        latest_version_created_at=updated_at,
        artifact=_get_artifact(current_package),
        authors=published_by["name"] if published_by else None,
    )


def _update_advisories(package: Package) -> None:
    pkg_advisories = advisories.get_vulnerabilities("cargo", package.name, package.version)
    if pkg_advisories:
        package.advisories = pkg_advisories


def complete_package(package: Package) -> Package:
    _update_advisories(package)
    cargo_package = get_cargo_package(package.name)
    if not cargo_package:
        return package

    versions = cargo_package["versions"]
    current_package = next(
        (version for version in versions if version["num"] == package.version),
        None,
    )

    _set_health_metadata(package, current_package, cargo_package)

    licenses = versions[0].get("license")
    if isinstance(licenses, str):
        package.licenses = validate_licenses([licenses])

    return package



from typing import NotRequired, TypedDict

import requests

from labels.internal.cache import dual_cache


class Author(TypedDict):
    name: str
    homepage: str
    email: NotRequired[str]


class Source(TypedDict):
    url: str
    type: str
    reference: str


class Dist(TypedDict):
    url: str
    type: str
    shasum: str
    reference: str


class PackagistPackageInfo(TypedDict, total=False):
    name: str
    description: str
    keywords: list[str]
    homepage: str
    version: str
    version_normalized: str
    license: list[str]
    authors: list[Author]
    source: Source
    dist: Dist
    type: str
    support: dict[str, str]
    funding: list[str]
    time: str
    autoload: dict[str, list[str]]
    extra: dict[str, dict[str, str]]
    require: dict[str, str]
    require_dev: dict[str, str]
    suggest: dict[str, str]
    conflict: dict[str, str]


class PackagistResponse(TypedDict, total=False):
    minified: str
    packages: dict[str, list[PackagistPackageInfo]]


@dual_cache
def get_composer_package(
    package_name: str,
    version: str | None = None,
) -> PackagistPackageInfo | None:
    base_url = f"https://repo.packagist.org/p2/{package_name}.json"
    response = requests.get(base_url, timeout=30)
    if response.status_code != 200:
        return None

    response_data: PackagistResponse = response.json()
    package_versions: list[PackagistPackageInfo] = response_data.get("packages", {}).get(
        package_name,
        [],
    )

    if version:
        for version_data in package_versions:
            if version_data.get("version") == version:
                return version_data
        return None

    if package_versions:
        return package_versions[0]

    return None



from packageurl import (
    PackageURL,
)

from labels.advisories import (
    roots as advisories,
)
from labels.internal.package_information.ruby import (
    RubyGemsPackage,
    get_gem_package,
)
from labels.model.core import (
    Artifact,
    Digest,
    HealthMetadata,
    Package,
)
from labels.pkg.cataloger.common import (
    infer_algorithm,
)
from labels.utils.licenses.validation import (
    validate_licenses,
)


def package_url(name: str, version: str) -> str:
    return PackageURL(
        type="gem",
        namespace="",
        name=name,
        version=version,
        qualifiers=None,
        subpath="",
    ).to_string()


def _get_artifact(current_package: RubyGemsPackage) -> Artifact | None:
    digest_value = current_package.get("sha") or None
    return Artifact(
        url=current_package["gem_uri"],
        integrity=Digest(
            algorithm=infer_algorithm(digest_value),
            value=digest_value,
        ),
    )


def _set_health_metadata(
    package: Package,
    gem_package: RubyGemsPackage,
    current_package: RubyGemsPackage | None,
) -> None:
    package.health_metadata = HealthMetadata(
        latest_version=gem_package["version"],
        latest_version_created_at=gem_package["version_created_at"],
        authors=gem_package["authors"],
        artifact=_get_artifact(current_package) if current_package else None,
    )


def _update_advisories(package: Package) -> None:
    pkg_advisories = advisories.get_vulnerabilities("gem", package.name, package.version)
    if pkg_advisories:
        package.advisories = pkg_advisories


def complete_package(package: Package) -> Package:
    _update_advisories(package)
    current_package = get_gem_package(package.name, package.version)
    gem_package = get_gem_package(package.name)
    if not gem_package:
        return package

    _set_health_metadata(package, gem_package, current_package)

    gem_licenses = gem_package["licenses"]
    if gem_licenses:
        package.licenses = validate_licenses(gem_licenses)
    return package



import asyncio

from labels.internal.file_resolver.container_image import ContainerImage
from labels.model.core import SbomConfig, SourceType
from labels.sources.directory_source import Directory
from labels.sources.docker import get_docker_image, get_image_context
from labels.sources.ecr import AwsRole, ecr_connection
from labels.utils.exceptions import UnexpectedSBOMSourceError


def _handle_directory(sbom_config: SbomConfig) -> Directory:
    return Directory(
        root=sbom_config.source,
        exclude=sbom_config.exclude,
    )


def _handle_docker(*, sbom_config: SbomConfig, is_daemon: bool = False) -> ContainerImage:
    docker_image = get_docker_image(
        sbom_config.source,
        username=sbom_config.docker_user,
        password=sbom_config.docker_password,
        daemon=is_daemon,
    )

    if not docker_image:
        error_msg = f"No image found for {sbom_config.source}"
        raise ValueError(error_msg)

    context = get_image_context(
        image=docker_image,
        username=sbom_config.docker_user,
        password=sbom_config.docker_password,
        daemon=is_daemon,
    )

    if context is None:
        error_msg = f"No context found for {docker_image}"
        raise ValueError(error_msg)

    return ContainerImage(
        img=docker_image,
        context=context,
        lazy=False,
    )


def _handle_ecr(sbom_config: SbomConfig) -> ContainerImage:
    if not sbom_config.aws_role:
        error_msg = "The AWS role wasn't defined"
        raise ValueError(error_msg)

    role = AwsRole(
        external_id=sbom_config.aws_external_id,
        role=sbom_config.aws_role,
    )

    token, image_metadata = asyncio.run(
        ecr_connection(role, sbom_config.source),
    )

    if not image_metadata:
        error_msg = f"No image found for {sbom_config.source}"
        raise ValueError(error_msg)

    context = get_image_context(
        image=image_metadata,
        aws_creds=f"AWS:{token}",
    )

    if context is None:
        error_msg = f"No context found for {image_metadata}"
        raise ValueError(error_msg)

    return ContainerImage(
        img=image_metadata,
        context=context,
        lazy=False,
    )


def resolve_sbom_source(sbom_config: SbomConfig) -> Directory | ContainerImage:
    match sbom_config.source_type:
        case SourceType.DIRECTORY:
            return _handle_directory(sbom_config)
        case SourceType.DOCKER:
            return _handle_docker(sbom_config=sbom_config)
        case SourceType.DOCKER_DAEMON:
            return _handle_docker(sbom_config=sbom_config, is_daemon=True)
        case SourceType.ECR:
            return _handle_ecr(sbom_config)
        case _:
            raise UnexpectedSBOMSourceError



import hashlib
from io import BufferedReader

from labels.model.core import Digest


def new_digests_from_file(file_object: BufferedReader, hashes: list[str]) -> list[Digest]:
    # Create hash objects
    hash_objects = [hashlib.new(hash_key, usedforsecurity=False) for hash_key in hashes]

    # Read file and update each hash object
    while chunk := file_object.read(4096):
        for hasher in hash_objects:
            hasher.update(chunk)

    # Prepare the result list of Digest objects
    if file_object.tell() == 0:  # Check if file size is zero
        return []

    return [
        Digest(
            algorithm=hash_name,
            value=hasher.hexdigest(),
        )
        for hash_name, hasher in zip(hashes, hash_objects, strict=False)
    ]



from packageurl import (
    PackageURL,
)

from labels.advisories import (
    roots as advisories,
)
from labels.model.core import Package


def package_url(name: str, version: str) -> str:
    return PackageURL(  # type: ignore
        type="conan",
        namespace="",
        name=name,
        version=version,
        qualifiers=None,
        subpath="",
    ).to_string()


def _update_advisories(package: Package) -> None:
    pkg_advisories = advisories.get_vulnerabilities("conan", package.name, package.version)
    if pkg_advisories:
        package.advisories = pkg_advisories


def complete_package(package: Package) -> Package:
    _update_advisories(package)

    return package



from typing import (
    TypedDict,
)

from labels.internal.package_information.api_interface import (
    make_get,
)


class DebianVersionInfo(TypedDict):
    area: str
    suites: list[str]
    version: str


class DebianPackageInfo(TypedDict):
    package: str
    path: str
    pathl: list[tuple[str, str]]
    suite: str
    type: str
    versions: list[DebianVersionInfo]


class StatInfo(TypedDict):
    perms: str
    size: int
    symlink_dest: str | None
    type: str


class ContentItem(TypedDict):
    hidden: bool
    name: str
    percent_encoded_name: str
    stat: StatInfo
    type: str


class MetricInfo(TypedDict):
    size: int


class PkgInfos(TypedDict):
    area: str
    copyright: bool
    ctags_count: int
    license: str
    metric: MetricInfo
    pts_link: str
    sloc: list[tuple[str, int]]
    suites: list[str]
    vcs_browser: str
    vcs_type: str


class DebianDirectoryInfo(TypedDict):
    content: list[ContentItem]
    directory: str
    package: str
    path: str
    pkg_infos: PkgInfos
    type: str
    version: str


def get_deb_package_version_list(
    package_name: str,
    release_code_name: str | None = None,
) -> list[DebianVersionInfo] | None:
    package_info: DebianPackageInfo | None = make_get(
        f"https://sources.debian.org/api/src/{package_name}/",
    )
    if not package_info or "error" in package_info:
        return None
    versions = package_info["versions"]
    if release_code_name:
        versions = [
            x for x in versions if (release_code_name in x["suites"] if "suites" in x else True)
        ]
    return versions


def get_deb_package(package_name: str, version: str) -> DebianDirectoryInfo | None:
    package_info: DebianDirectoryInfo | None = make_get(
        f"https://sources.debian.org/api/src/{package_name}/{version}/",
    )
    if not package_info:
        return None

    return package_info


def get_deb_snapshot(package_name: str, version: str) -> str | None:
    url = f"https://snapshot.debian.org/package/{package_name}/{version}/"
    result = make_get(
        url,
        content=True,
    )
    if not result or not isinstance(result, str):
        return None
    return result



import logging
from copy import (
    deepcopy,
)

from pydantic import (
    BaseModel,
    ValidationError,
)

from labels.artifact.relationship import (
    Relationship,
)
from labels.file.location_read_closer import (
    LocationReadCloser,
)
from labels.file.resolver import (
    Resolver,
)
from labels.model.core import (
    Language,
    Package,
    PackageType,
)
from labels.pkg.cataloger.generic.parser import (
    Environment,
)
from labels.pkg.cataloger.java.model import (
    JavaArchive,
    JavaPomProject,
)
from labels.pkg.cataloger.java.package import (
    package_url,
)
from labels.utils.strings import format_exception

LOGGER = logging.getLogger(__name__)


class LockFileDependency(BaseModel):
    group: str
    name: str
    version: str
    line: int | None = None


def parse_gradle_lockfile(
    _resolver: Resolver | None,
    __: Environment | None,
    reader: LocationReadCloser,
) -> tuple[list[Package], list[Relationship]]:
    dependencies: list[LockFileDependency] = []
    packages: list[Package] = []
    for line_number, line in enumerate(reader.read_closer.readlines(), 1):
        if "=" in line and ":" in line:  # To ensure it's a dependency line
            dependency_part = line.split("=")[0]
            group, name, version = dependency_part.split(":")
            dependencies.append(
                LockFileDependency(group=group, name=name, version=version, line=line_number),
            )

    for dependency in dependencies:
        name = dependency.name
        version = dependency.version

        if not name or not version:
            continue

        location = deepcopy(reader.location)
        if location.coordinates:
            location.coordinates.line = dependency.line

        archive = JavaArchive(
            pom_project=JavaPomProject(
                group_id=dependency.group,
                name=name,
                artifact_id=name,
                version=version,
            ),
        )

        try:
            packages.append(
                Package(
                    name=name,
                    version=version,
                    locations=[location],
                    language=Language.JAVA,
                    type=PackageType.JavaPkg,
                    metadata=archive,
                    p_url=package_url(name, version, archive),
                    licenses=[],
                ),
            )
        except ValidationError as ex:
            LOGGER.warning(
                "Malformed package. Required fields are missing or data types are incorrect.",
                extra={
                    "extra": {
                        "exception": format_exception(str(ex)),
                        "location": location.path(),
                    },
                },
            )
            continue

    return packages, []



from copy import (
    deepcopy,
)
from typing import (
    cast,
)

from labels.artifact.relationship import (
    Relationship,
)
from labels.file.location_read_closer import (
    LocationReadCloser,
)
from labels.file.resolver import (
    Resolver,
)
from labels.internal.collection.types import (
    IndexedDict,
    IndexedList,
    ParsedValue,
)
from labels.internal.collection.yaml import (
    parse_yaml_with_tree_sitter,
)
from labels.model.core import (
    Language,
    Package,
    PackageType,
)
from labels.pkg.cataloger.generic.parser import (
    Environment,
)
from labels.pkg.cataloger.github.package import (
    package_url,
)


def _get_deps(jobs: IndexedDict[str, ParsedValue]) -> list[tuple[str, int]]:
    deps: list[tuple[str, int]] = []
    for job in jobs.values():
        if not isinstance(job, IndexedDict):
            continue
        steps = job.get("steps")
        if not isinstance(steps, IndexedList):
            continue
        for step in steps:
            if not isinstance(step, IndexedDict):
                continue
            uses = step.get("uses")
            if not isinstance(uses, str):
                continue
            deps.append((uses, step.get_key_position("uses").start.line))
    return deps


def parse_github_actions_deps(
    _: Resolver | None,
    __: Environment | None,
    reader: LocationReadCloser,
) -> tuple[list[Package], list[Relationship]]:
    packages: list[Package] = []
    parsed_content = cast(
        IndexedDict[str, ParsedValue],
        parse_yaml_with_tree_sitter(reader.read_closer.read()),
    )
    if not parsed_content:
        return packages, []
    jobs = parsed_content.get("jobs")
    if not jobs or not isinstance(jobs, IndexedDict):
        return packages, []

    for dep, line_number in _get_deps(jobs):
        dep_info = dep.rsplit("@", 1)
        location = deepcopy(reader.location)
        if location.coordinates:
            location.coordinates.line = line_number
        if len(dep_info) == 2:
            packages.append(
                Package(
                    name=dep_info[0],
                    version=dep_info[1],
                    language=Language.UNKNOWN_LANGUAGE,
                    licenses=[],
                    locations=[location],
                    type=PackageType.GithubActionPkg,
                    metadata=None,
                    p_url=package_url(dep_info[0], dep_info[1]),
                ),
            )
    return packages, []



from __future__ import (
    annotations,
)

import functools
import json
import logging
from concurrent.futures import (
    ThreadPoolExecutor,
)
from multiprocessing import (
    cpu_count,
)
from typing import (
    cast,
)

from packageurl import (
    PackageURL,
)

from labels.advisories.images.database import (
    DATABASE,
)
from labels.advisories.images.namespace import (
    matches_namespace,
)
from labels.advisories.images.version import (
    matches_version,
)
from labels.model.core import (
    Advisory,
    Language,
    Package,
)

LOGGER = logging.getLogger(__name__)


VALID_NAMESPACES = [
    "alpine:distro:alpine:3.10",
    "alpine:distro:alpine:3.11",
    "alpine:distro:alpine:3.12",
    "alpine:distro:alpine:3.13",
    "alpine:distro:alpine:3.14",
    "alpine:distro:alpine:3.15",
    "alpine:distro:alpine:3.16",
    "alpine:distro:alpine:3.17",
    "alpine:distro:alpine:3.18",
    "alpine:distro:alpine:3.19",
    "alpine:distro:alpine:3.2",
    "alpine:distro:alpine:3.20",
    "alpine:distro:alpine:3.3",
    "alpine:distro:alpine:3.4",
    "alpine:distro:alpine:3.5",
    "alpine:distro:alpine:3.6",
    "alpine:distro:alpine:3.7",
    "alpine:distro:alpine:3.8",
    "alpine:distro:alpine:3.9",
    "alpine:distro:alpine:edge",
    "amazon:distro:amazonlinux:2",
    "amazon:distro:amazonlinux:2022",
    "amazon:distro:amazonlinux:2023",
    "chainguard:distro:chainguard:rolling",
    "debian:distro:debian:11",
    "debian:distro:debian:12",
    "debian:distro:debian:13",
    "debian:distro:debian:unstable",
    "github:language:dart",
    "github:language:dotnet",
    "github:language:go",
    "github:language:java",
    "github:language:javascript",
    "github:language:php",
    "github:language:python",
    "github:language:ruby",
    "github:language:rust",
    "github:language:swift",
    "mariner:distro:mariner:1.0",
    "mariner:distro:mariner:2.0",
    "nvd:cpe",
    "oracle:distro:oraclelinux:5",
    "oracle:distro:oraclelinux:6",
    "oracle:distro:oraclelinux:7",
    "oracle:distro:oraclelinux:8",
    "oracle:distro:oraclelinux:9",
    "redhat:distro:redhat:5",
    "redhat:distro:redhat:6",
    "redhat:distro:redhat:7",
    "redhat:distro:redhat:8",
    "redhat:distro:redhat:9",
    "sles:distro:sles:11",
    "sles:distro:sles:11.1",
    "sles:distro:sles:11.2",
    "sles:distro:sles:11.3",
    "sles:distro:sles:11.4",
    "sles:distro:sles:12",
    "sles:distro:sles:12.1",
    "sles:distro:sles:12.2",
    "sles:distro:sles:12.3",
    "sles:distro:sles:12.4",
    "sles:distro:sles:12.5",
    "sles:distro:sles:15",
    "sles:distro:sles:15.1",
    "sles:distro:sles:15.2",
    "sles:distro:sles:15.3",
    "sles:distro:sles:15.4",
    "sles:distro:sles:15.5",
    "sles:distro:sles:15.6",
    "ubuntu:distro:ubuntu:12.04",
    "ubuntu:distro:ubuntu:12.10",
    "ubuntu:distro:ubuntu:13.04",
    "ubuntu:distro:ubuntu:14.04",
    "ubuntu:distro:ubuntu:14.10",
    "ubuntu:distro:ubuntu:15.04",
    "ubuntu:distro:ubuntu:15.10",
    "ubuntu:distro:ubuntu:16.04",
    "ubuntu:distro:ubuntu:16.10",
    "ubuntu:distro:ubuntu:17.04",
    "ubuntu:distro:ubuntu:17.10",
    "ubuntu:distro:ubuntu:18.04",
    "ubuntu:distro:ubuntu:18.10",
    "ubuntu:distro:ubuntu:19.04",
    "ubuntu:distro:ubuntu:19.10",
    "ubuntu:distro:ubuntu:20.04",
    "ubuntu:distro:ubuntu:20.10",
    "ubuntu:distro:ubuntu:21.04",
    "ubuntu:distro:ubuntu:21.10",
    "ubuntu:distro:ubuntu:22.04",
    "ubuntu:distro:ubuntu:22.10",
    "ubuntu:distro:ubuntu:23.04",
    "ubuntu:distro:ubuntu:23.10",
    "ubuntu:distro:ubuntu:24.04",
    "wolfi:distro:wolfi:rolling",
]


AdvisoryRow = tuple[str, str, str, str, str, float | None, float | None, str, str]


def _format_advisory(row: AdvisoryRow) -> Advisory:
    return Advisory(
        cpes=cast(list[str], json.loads(row[0] or "[]")),
        description=row[4] or None,
        epss=row[5] or 0.0,
        id=row[1],
        namespace=row[2],
        percentile=row[6] or 0.0,
        severity=row[7],
        urls=[url for url in cast(list[str], json.loads(row[8] or "[]")) if url],
        version_constraint=row[3] or None,
        cvss3=None,
        cvss4=None,
    )


def get_namespace_sufix(
    package: Package,
    *,
    distro_id: str | None = None,
    distro_version: str | None = None,
) -> list[str] | None:
    distro_result = []
    if (not distro_id or not distro_version) and (
        package.p_url
        and (
            pacakge_url := PackageURL.from_string(  # type: ignore[misc]
                package.p_url,
            )
        )
        and isinstance(pacakge_url.qualifiers, dict)  # type: ignore[misc]
    ):
        qualifiers = pacakge_url.qualifiers  # type: ignore[misc]
        distro_id = qualifiers.get("distro_id")
        distro_version = qualifiers.get("distro_version_id")

    if distro_id and distro_version:
        distro_version = ".".join(distro_version.split(".")[0:2])
        namespace = f"{distro_id}:distro:{distro_id}:{distro_version}"
        if namespace in VALID_NAMESPACES:
            distro_result.append(namespace)
        elif distro_id == "alpine":
            # If the distro version does not have the x.y.z format, edge should
            distro_result.append("alpine:distro:alpine:edge")
        distro_result.append("nvd:cpe")
    if distro_result:
        return distro_result

    if package.language != Language.UNKNOWN_LANGUAGE:
        return [
            "nvd:cpe",
            f"github:language:{package.language.value}",
        ]

    return None


def remove_duplicates(advisories: list[Advisory]) -> list[Advisory]:
    advisory_dict: dict[str, Advisory] = {}

    for advisory in advisories:
        if advisory.id in advisory_dict:
            if advisory.get_info_count() > advisory_dict[advisory.id].get_info_count():
                advisory_dict[advisory.id] = advisory
        else:
            advisory_dict[advisory.id] = advisory

    return list(advisory_dict.values())


def _handle_raw_advisory(package: Package, raw_advisory: AdvisoryRow) -> Advisory | None:
    advisory = _format_advisory(raw_advisory)
    if matches_namespace(package, advisory) and matches_version(package, advisory):
        return advisory
    return None


def _get_matching_advisories(
    package: Package,
    *,
    distro_id: str | None = None,
    distro_version: str | None = None,
) -> list[Advisory]:
    connection = DATABASE.get_connection()
    if connection is None:
        return []
    advisories_raw = []
    cursor = connection.cursor()
    if namespace_sufix := get_namespace_sufix(
        package,
        distro_id=distro_id,
        distro_version=distro_version,
    ):
        for namespace in namespace_sufix:
            if namespace not in VALID_NAMESPACES:
                LOGGER.warning(
                    "Invalid namespace",
                    extra={"extra": {"namespace": namespace}},
                )
                continue
            cursor.execute(
                """
                SELECT
                    vuln.cpes,
                    vuln.id,
                    vuln.namespace,
                    vuln.version_constraint,
                    vuln_meta.description,
                    vuln_meta.epss,
                    vuln_meta.percentile,
                    vuln_meta.severity,
                    vuln_meta.urls
                FROM
                    vulnerability vuln
                INNER JOIN vulnerability_metadata vuln_meta
                    ON vuln.id = vuln_meta.id
                    AND vuln.namespace = vuln_meta.namespace
                WHERE vuln.package_name = ?
                AND vuln.namespace = ?
                """,
                (package.name.lower(), namespace),
            )
            advisories_raw.extend(cast(list[AdvisoryRow], cursor.fetchall()))
    else:
        cursor.execute(
            """
            SELECT
                vuln.cpes,
                vuln.id,
                vuln.namespace,
                vuln.version_constraint,
                vuln_meta.description,
                vuln_meta.epss,
                vuln_meta.percentile,
                vuln_meta.severity,
                vuln_meta.urls
            FROM
                vulnerability vuln
            INNER JOIN vulnerability_metadata vuln_meta
                ON vuln.id = vuln_meta.id
                AND vuln.namespace = vuln_meta.namespace
            WHERE vuln.package_name = ?
            """,
            (package.name.lower(),),
        )
        advisories_raw.extend(cast(list[AdvisoryRow], cursor.fetchall()))
    with ThreadPoolExecutor(max_workers=cpu_count()) as executor:
        result = list(
            filter(
                None,
                executor.map(
                    functools.partial(_handle_raw_advisory, package),
                    advisories_raw,
                ),
            ),
        )
    return remove_duplicates(result)


def get_package_advisories(
    package: Package,
    *,
    distro_id: str | None = None,
    distro_version: str | None = None,
) -> list[Advisory]:
    try:
        return _get_matching_advisories(package, distro_id=distro_id, distro_version=distro_version)
    except Exception:
        LOGGER.exception(
            "Unable to get advisories for package %s",
            package.name,
        )
        return []


def _get_advisories_test() -> list[Advisory]:
    DATABASE.initialize()
    connection = DATABASE.get_connection()
    if connection is None:
        return []
    cursor = connection.cursor()
    cursor.execute(
        """
        SELECT
            vuln.cpes,
            vuln.id,
            vuln.namespace,
            vuln.version_constraint
        FROM
            vulnerability vuln
        WHERE vuln.id = ?
        """,
        ("CVE-2015-6096",),
    )
    return cast(list[Advisory], cursor.fetchall())


if __name__ == "__main__":
    _get_advisories_test()



import logging
from copy import (
    deepcopy,
)
from typing import (
    cast,
)

from labels.artifact.relationship import (
    Relationship,
    RelationshipType,
)
from labels.file.location import (
    Location,
)
from labels.file.location_read_closer import (
    LocationReadCloser,
)
from labels.file.resolver import (
    Resolver,
)
from labels.internal.collection.json import (
    parse_json_with_tree_sitter,
)
from labels.internal.collection.types import (
    IndexedDict,
    IndexedList,
    ParsedValue,
)
from labels.model.core import (
    Package,
)
from labels.pkg.cataloger.generic.parser import (
    Environment,
)
from labels.pkg.cataloger.php.model import PhpComposerInstalledEntry, PhpComposerLockEntry
from labels.pkg.cataloger.php.package import (
    new_package_from_composer,
)

LOGGER = logging.getLogger(__name__)
EMPTY_LIST: IndexedList[ParsedValue] = IndexedList()


def _extract_packages(
    package_json: IndexedDict[str, ParsedValue],
    location: Location,
) -> list[Package]:
    packages = []
    if isinstance(package_json, IndexedDict) and "dev-package-names" in package_json:
        dev_packages = cast(IndexedList[ParsedValue], package_json["dev-package-names"])
    else:
        dev_packages = EMPTY_LIST

    packages_list = (
        package_json["packages"] if isinstance(package_json, IndexedDict) else package_json
    )

    for package in packages_list if isinstance(packages_list, IndexedList) else EMPTY_LIST:
        if not isinstance(package, IndexedDict):
            continue
        pkg_item = new_package_from_composer(package, deepcopy(location), is_dev=False)
        if pkg_item:
            pkg_item.is_dev = pkg_item.name in dev_packages
            packages.append(pkg_item)

    return packages


def _extract_relationships(packages: list[Package]) -> list[Relationship]:
    relationships = []
    for package in packages:
        if not isinstance(package.metadata, PhpComposerInstalledEntry | PhpComposerLockEntry):
            continue
        package_metadata = package.metadata.require
        dependencies = list(package_metadata.keys()) if package_metadata else []
        for dep_name in dependencies:
            package_dep = next((x for x in packages if x.name == dep_name), None)
            if package_dep:
                relationships.append(
                    Relationship(
                        from_=package,
                        to_=package_dep,
                        type=RelationshipType.DEPENDENCY_OF_RELATIONSHIP,
                    ),
                )

    return relationships


def parse_installed_json(
    _: Resolver | None,
    __: Environment | None,
    reader: LocationReadCloser,
) -> tuple[list[Package], list[Relationship]]:
    package_json = cast(
        IndexedDict[str, ParsedValue],
        parse_json_with_tree_sitter(reader.read_closer.read()),
    )

    packages = _extract_packages(package_json, reader.location)
    relationships = _extract_relationships(packages)

    return packages, relationships



import logging
from contextlib import (
    suppress,
)

from labels.pkg.cataloger.redhat.rpmdb import (
    berkeley,
    sqlite,
)
from labels.pkg.cataloger.redhat.rpmdb.entry import (
    header_import,
)
from labels.pkg.cataloger.redhat.rpmdb.ndb import (
    RpmNDB,
)
from labels.pkg.cataloger.redhat.rpmdb.package import (
    PackageInfo,
    get_nevra,
)
from labels.pkg.cataloger.redhat.rpmdb.rpmdb_interface import (
    RpmDBInterface,
)
from labels.utils.exceptions import (
    InvalidDBFormatError,
    InvalidMetadataError,
)

LOGGER = logging.getLogger(__name__)


class RpmDB:
    def __init__(self, database: RpmDBInterface) -> None:
        self.database = database

    def list_packages(
        self,
    ) -> list[PackageInfo]:
        packages: list[PackageInfo] = []
        for entry in self.database.read():
            try:
                index_entries = header_import(entry)
            except ValueError:
                LOGGER.exception("Failed to import header")
                continue
            if index_entries:
                try:
                    package = get_nevra(index_entries)
                except ValueError:
                    LOGGER.exception("Failed to get nevra from index entries")
                    continue
                packages.append(package)
        return packages


def open_db(file_path: str) -> RpmDB | None:
    """Attempt to open an RPM database from the specified file path and returns an RpmDB instance.

    If the database is invalid or the metadata cannot be
    validated, None is returned.

    The function first tries to open the database as an SQLite database, and
    if that fails, it attempts to open it as a Berkeley DB. If both attempts
    fail, None is returned.

    :param file_path: The path to the RPM database file.
    :type file_path: str
    :return: An RpmDB instance if the database is valid, otherwise None.
    :rtype: RpmDB | None
    """
    with suppress(InvalidDBFormatError):
        return RpmDB(sqlite.open_sqlite(file_path))

    with suppress(InvalidDBFormatError):
        return RpmDB(RpmNDB.open(file_path))

    try:
        return RpmDB(berkeley.open_berkeley(file_path))
    except InvalidDBFormatError:
        pass
    except (ValueError, InvalidMetadataError):
        LOGGER.exception("Failed to open RPM database")

    return None



import logging
from typing import (
    TYPE_CHECKING,
)

from cyclonedx.model.bom import (
    Bom,
)
from cyclonedx.output import (
    make_outputter,
)
from cyclonedx.schema import (
    OutputFormat,
    SchemaVersion,
)
from cyclonedx.validation import (
    make_schemabased_validator,
)
from cyclonedx.validation.json import (
    JsonStrictValidator,
)

from labels.utils.exceptions import (
    CycloneDXValidationError,
)

LOGGER = logging.getLogger(__name__)

if TYPE_CHECKING:
    from cyclonedx.output.json import (
        Json as JsonOutputter,
    )
    from cyclonedx.output.xml import (
        Xml as XmlOutputter,
    )
    from cyclonedx.validation.xml import (
        XmlValidator,
    )


def format_cyclone_json(bom: Bom, output: str) -> None:
    file_path = f"{output}.json"
    json_output: JsonOutputter = make_outputter(
        bom=bom,
        output_format=OutputFormat.JSON,
        schema_version=SchemaVersion.V1_6,
    )
    serialized_json = json_output.output_as_string()

    json_validator = JsonStrictValidator(SchemaVersion.V1_6)
    validation_error = json_validator.validate_str(serialized_json)

    if validation_error:
        raise CycloneDXValidationError(validation_error)

    LOGGER.info(
        "🆗 Valid CYCLONEDX JSON format, generating output file at %s",
        file_path,
    )
    json_output.output_to_file(file_path, allow_overwrite=True, indent=2)
    LOGGER.info("✅ Output file successfully generated")


def format_cyclone_xml(bom: Bom, output: str) -> None:
    file_path = f"{output}.xml"
    xml_outputter: XmlOutputter = make_outputter(
        bom=bom,
        output_format=OutputFormat.XML,
        schema_version=SchemaVersion.V1_6,
    )
    serialized_xml = xml_outputter.output_as_string()

    xml_validator: XmlValidator = make_schemabased_validator(
        output_format=OutputFormat.XML,
        schema_version=SchemaVersion.V1_6,
    )
    validation_error = xml_validator.validate_str(serialized_xml)

    if validation_error:
        raise CycloneDXValidationError(validation_error)

    LOGGER.info("🆗 Valid CYCLONEDX XML format, generating output file at %s", file_path)
    xml_outputter.output_to_file(file_path, allow_overwrite=True, indent=2)
    LOGGER.info("✅ Output file successfully generated")



import logging
import os
from itertools import (
    zip_longest,
)
from typing import cast

from pydantic import (
    ValidationError,
)

from labels.artifact.relationship import (
    Relationship,
)
from labels.file.location_read_closer import (
    LocationReadCloser,
)
from labels.file.resolver import (
    Resolver,
)
from labels.model.core import (
    Digest,
    Language,
    Package,
    PackageType,
)
from labels.pkg.cataloger.generic.parser import (
    Environment,
)
from labels.pkg.cataloger.redhat.package import (
    RpmDBEntry,
    RpmFileRecord,
    package_url,
)
from labels.pkg.cataloger.redhat.rpmdb import (
    open_db,
)
from labels.pkg.cataloger.redhat.rpmdb.package import (
    PackageInfo,
)
from labels.utils.strings import format_exception

LOGGER = logging.getLogger(__name__)


def to_int(value: int | None, default: int = 0) -> int:
    return int(value) if isinstance(value, int) else default


def extract_rmp_file_records(resolver: Resolver, entry: PackageInfo) -> list[RpmFileRecord]:
    records: list[RpmFileRecord] = []
    file_attributes = cast(
        tuple[tuple[str, int, str, int, int, int, str, str], ...],
        zip_longest(
            entry.base_names,
            entry.dir_indexes,
            entry.file_digests,
            entry.file_flags,
            entry.file_modes,
            entry.file_sizes,
            entry.user_names,
            entry.group_names,
        ),
    )

    records.extend(
        record
        for attrs in file_attributes
        if (record := create_rpm_file_record(resolver, entry, attrs))
    )

    return records


def create_rpm_file_record(
    resolver: Resolver,
    entry: PackageInfo,
    attrs: tuple[str, int, str, int, int, int, str, str],
) -> RpmFileRecord | None:
    (
        base_name,
        dir_index,
        file_digest,
        file_flag,
        file_mode,
        file_size,
        file_username,
        file_groupname,
    ) = attrs

    if not base_name or not isinstance(dir_index, int):
        return None

    file_path = os.path.join(str(entry.dir_names[dir_index]), str(base_name))
    file_location = resolver.files_by_path(file_path)

    if not file_location:
        return None

    return RpmFileRecord(
        path=file_path,
        mode=to_int(file_mode, default=0),
        size=to_int(file_size, default=0),
        digest=Digest(
            algorithm="md5" if file_digest else None,
            value=str(file_digest) if file_digest else None,
        ),
        username=str(file_username),
        flags=str(file_flag),
        group_name=str(file_groupname) if file_groupname else None,
    )


def parse_rpm_db(
    resolver: Resolver,
    env: Environment,
    reader: LocationReadCloser,
) -> tuple[list[Package], list[Relationship]]:
    packages: list[Package] = []

    if not reader.location.coordinates:
        return packages, []
    database = open_db(reader.location.coordinates.real_path)
    if not database:
        return packages, []

    for entry in database.list_packages():
        name = entry.name
        version = entry.version

        if not name or not version:
            continue

        metadata = RpmDBEntry(
            id_="",
            name=name,
            version=version,
            epoch=entry.epoch,
            arch=entry.arch,
            release=entry.release,
            source_rpm=entry.source_rpm,
            vendor=entry.vendor,
            size=entry.size,
            modularitylabel=entry.modularity_label,
            files=extract_rmp_file_records(resolver, entry),
        )
        try:
            packages.append(
                Package(
                    name=name,
                    version=version,
                    locations=[reader.location],
                    language=Language.UNKNOWN_LANGUAGE,
                    licenses=[entry.license],
                    type=PackageType.RpmPkg,
                    metadata=metadata,
                    p_url=package_url(
                        name=name,
                        arch=entry.arch,
                        epoch=entry.epoch,
                        source_rpm=entry.source_rpm,
                        version=version,
                        release=entry.release,
                        distro=env.linux_release,
                    ),
                ),
            )
        except ValidationError as ex:
            LOGGER.warning(
                "Malformed package. Required fields are missing or data types are incorrect.",
                extra={
                    "extra": {
                        "exception": format_exception(str(ex)),
                        "location": reader.location.path(),
                    },
                },
            )
            continue

    return packages, []



import logging
from copy import (
    deepcopy,
)
from typing import (
    cast,
)

from pydantic import (
    ValidationError,
)

from labels.artifact.relationship import (
    Relationship,
)
from labels.file.dependency_type import (
    DependencyType,
)
from labels.file.location_read_closer import (
    LocationReadCloser,
)
from labels.file.resolver import (
    Resolver,
)
from labels.file.scope import (
    Scope,
)
from labels.internal.collection.json import (
    parse_json_with_tree_sitter,
)
from labels.internal.collection.types import (
    IndexedDict,
    ParsedValue,
)
from labels.model.core import (
    Language,
    Package,
    PackageType,
)
from labels.pkg.cataloger.generic.parser import (
    Environment,
)
from labels.pkg.cataloger.javascript.model import (
    NpmPackage,
)
from labels.pkg.cataloger.javascript.package import (
    package_url,
)
from labels.utils.strings import format_exception

LOGGER = logging.getLogger(__name__)


def _create_package(
    package_json: IndexedDict[str, ParsedValue],
    reader: LocationReadCloser,
    package_name: str,
    specifier: str,
    *,
    is_dev: bool,
) -> Package | None:
    current_location = deepcopy(reader.location)
    current_location.scope = Scope.DEV if is_dev else Scope.PROD
    dependencies_key = "devDependencies" if is_dev else "dependencies"
    pkg: IndexedDict[str, ParsedValue] = cast(
        IndexedDict[str, ParsedValue],
        package_json[dependencies_key],
    )
    if current_location.coordinates:
        current_location.coordinates.line = pkg.get_key_position(package_name).start.line
        current_location.dependency_type = DependencyType.DIRECT
    try:
        return Package(
            name=package_name,
            version=specifier,
            type=PackageType.NpmPkg,
            language=Language.JAVASCRIPT,
            licenses=[],
            locations=[current_location],
            p_url=package_url(package_name, specifier),
            metadata=NpmPackage(
                name=package_name,
                version=specifier,
                is_dev=is_dev,
            ),
            is_dev=is_dev,
        )
    except ValidationError as ex:
        LOGGER.warning(
            "Malformed package. Required fields are missing or data types are incorrect.",
            extra={
                "extra": {
                    "exception": format_exception(str(ex)),
                    "location": current_location.path(),
                },
            },
        )
        return None


def parse_package_json(
    _: Resolver | None,
    __: Environment | None,
    reader: LocationReadCloser,
) -> tuple[list[Package], list[Relationship]]:
    package_json: IndexedDict[str, ParsedValue] = cast(
        IndexedDict[str, ParsedValue],
        parse_json_with_tree_sitter(reader.read_closer.read()),
    )
    deps: ParsedValue = package_json.get("dependencies", IndexedDict())
    if not isinstance(deps, IndexedDict):
        LOGGER.warning("No deps found in package JSON")
        return ([], [])

    packages = []
    for package_name, specifier in deps.items():
        if not package_name or not specifier:
            continue

        package = _create_package(
            package_json,
            reader,
            package_name,
            str(specifier),
            is_dev=False,
        )
        if package:
            packages.append(package)

    dev_deps: ParsedValue = package_json.get("devDependencies", IndexedDict())
    if not isinstance(dev_deps, IndexedDict):
        LOGGER.warning("No dev deps found in package JSON")
        return ([], [])
    for package_name, specifier in dev_deps.items():
        if not package_name or not specifier:
            continue

        package = _create_package(
            package_json,
            reader,
            package_name,
            str(specifier),
            is_dev=True,
        )
        if package:
            packages.append(package)

    return packages, []



import logging
from collections.abc import (
    Iterator,
)
from copy import deepcopy
from typing import cast

from more_itertools import flatten
from packageurl import PackageURL
from pydantic import ValidationError

from labels.artifact.relationship import Relationship, RelationshipType
from labels.file.location_read_closer import LocationReadCloser
from labels.file.resolver import Resolver
from labels.internal.collection.json import parse_json_with_tree_sitter
from labels.internal.collection.types import IndexedDict, ParsedValue
from labels.model.core import Language, Package, PackageType
from labels.pkg.cataloger.generic.parser import Environment
from labels.utils.strings import format_exception

LOGGER = logging.getLogger(__name__)

EMPTY_DICT: IndexedDict[str, ParsedValue] = IndexedDict()


def _get_relationships(
    packages: list[Package],
    dependencies: dict[str, ParsedValue],
) -> list[Relationship]:
    result: list[Relationship] = []
    for package_name, depens in dependencies.items():
        if not isinstance(depens, IndexedDict):
            continue
        deps_gen = (
            p
            for p in packages
            for dep_name in depens
            if dep_name == p.name and isinstance(depens, IndexedDict)
        )
        current_package = next((p for p in packages if p.name == package_name), None)
        if current_package is not None:
            result.extend(
                Relationship(
                    from_=p,
                    to_=current_package,
                    type=RelationshipType.DEPENDENCY_OF_RELATIONSHIP,
                )
                for p in deps_gen
            )
    return result


def parse_dotnet_deps_json(
    _resolver: Resolver | None,
    _env: Environment | None,
    reader: LocationReadCloser,
) -> tuple[list[Package], list[Relationship]]:
    package_json = cast(
        IndexedDict[str, ParsedValue],
        parse_json_with_tree_sitter(reader.read_closer.read()),
    )
    packages: list[Package] = []
    targets: ParsedValue = package_json.get("targets", IndexedDict())
    dependencies: dict[str, ParsedValue] = {}
    if not isinstance(targets, IndexedDict):
        LOGGER.warning("No targets found in package JSON")
        return ([], [])
    for package_key, package_value in cast(
        Iterator[IndexedDict[str, ParsedValue]],
        flatten(x.items() for x in targets.values() if isinstance(x, IndexedDict)),
    ):
        if not isinstance(package_key, str) or "/" not in package_key:
            continue
        package_name, version = package_key.split("/", 1)
        if not isinstance(package_value, IndexedDict):
            continue
        location = deepcopy(reader.location)
        pos = package_value.position
        if hasattr(location, "coordinates") and location.coordinates:
            location.coordinates.line = pos.start.line
        dependencies[package_name] = package_value.get("dependencies", EMPTY_DICT)
        try:
            pkg = Package(
                name=package_name,
                version=version,
                locations=[location],
                licenses=[],
                type=PackageType.DotnetPkg,
                language=Language.DOTNET,
                metadata=None,
                p_url=PackageURL(
                    type="nuget",
                    namespace="",
                    name=package_name,
                    version=version,
                    qualifiers={},
                    subpath="",
                ).to_string(),
            )
            packages.append(pkg)
        except ValidationError as ex:
            LOGGER.warning(
                "Malformed package. Required fields are missing or data types are incorrect.",
                extra={
                    "extra": {
                        "exception": format_exception(str(ex)),
                        "location": location.path(),
                    },
                },
            )
            continue

    relationships: list[Relationship] = _get_relationships(packages, dependencies)
    return packages, relationships



import logging
import re
from contextlib import (
    suppress,
)
from copy import (
    deepcopy,
)
from pathlib import Path
from typing import (
    TextIO,
)

from bs4 import (
    BeautifulSoup,
    NavigableString,
    Tag,
)
from pydantic import (
    BaseModel,
    ValidationError,
)

from labels.artifact.relationship import (
    Relationship,
)
from labels.file.dependency_type import (
    DependencyType,
)
from labels.file.location import (
    Location,
)
from labels.file.location_read_closer import (
    LocationReadCloser,
)
from labels.file.resolver import (
    Resolver,
)
from labels.model.core import (
    Language,
    Package,
    PackageType,
)
from labels.pkg.cataloger.generic.parser import (
    Environment,
)
from labels.pkg.cataloger.java.maven_repo_utils import (
    recursively_find_versions_from_parent_pom,
)
from labels.pkg.cataloger.java.model import (
    JavaArchive,
    JavaPomParent,
    JavaPomProject,
    JavaPomProperties,
)
from labels.pkg.cataloger.java.package import (
    package_url,
)
from labels.utils.strings import format_exception

LOGGER = logging.getLogger(__name__)


def get_pom_project(content: TextIO) -> Tag | None:
    if (
        (root := BeautifulSoup(content, features="html.parser"))
        and (pom_project := root.project)
        and (xmlns := root.project.get("xmlns"))
        and str(xmlns) == "http://maven.apache.org/POM/4.0.0"
    ):
        return pom_project
    return None


class ParsedPomProject(BaseModel):
    java_pom_project: JavaPomProject
    licenses: list[str]


def extract_bracketed_text(item: str) -> str:
    match = re.search(r"\$\{([^}]+)\}", item)
    if match:
        return match.group(1)
    return ""


def _get_text(parent: Tag | NavigableString, name: str) -> str | None:
    element = parent.find_next(name)
    if element:
        return element.get_text()
    return None


def _get_parent_info(project: Tag) -> dict[str, str] | None:
    if (
        (parent := project.find_next("parent"))
        and (parent_groupid_node := parent.find_next("groupid"))
        and (parent_artifactid_node := parent.find_next("artifactid"))
        and (parent_version_node := parent.find_next("version"))
    ):
        return {
            "group": parent_groupid_node.get_text(),
            "artifact": parent_artifactid_node.get_text(),
            "version": parent_version_node.get_text(),
        }
    return None


def _is_parent_pom(root_pom_project: Tag, parent_info: dict[str, str]) -> bool:
    group = root_pom_project.find("groupid", recursive=False)
    artifact = root_pom_project.find("artifactid", recursive=False)
    version = root_pom_project.find("version", recursive=False)
    if not (group and artifact and version):
        return False
    return (
        group.get_text() == parent_info["group"]
        and artifact.get_text() == parent_info["artifact"]
        and version.get_text() == parent_info["version"]
    )


def _is_module_parent(parent_pom: Tag, pom_module: str) -> bool:
    base_module = Path(pom_module).parent.name
    for modules in parent_pom.find_all("modules"):
        for module in modules.find_all("module"):
            mod_name = module.get_text()
            if mod_name == base_module:
                return True
    return False


def _get_properties(project_pom: Tag) -> dict[str, str]:
    return {
        _property.name.lower(): _property.get_text()
        for properties in project_pom.find_all("properties", limit=2)
        for _property in properties.children
        if isinstance(_property, Tag)
    }


def _get_deps_management(pom_tree: Tag) -> dict[str, str]:
    deps_info: dict[str, str] = {}
    for manage in pom_tree.find_all("dependencymanagement"):
        for dependency in manage.find_all("dependency", recursive=True):
            if not (dependency.groupid and dependency.artifactid and dependency.version):
                continue

            group = dependency.groupid.get_text()
            artifact = dependency.artifactid.get_text()
            version = dependency.version.get_text()
            deps_info[f"{group}:{artifact}"] = version
    return deps_info


def _evaluate_pom_files_in_project(
    resolver: Resolver,
    parent_info: dict[str, str],
    current_pom_file_path: str,
    current_pom_project: Tag,
) -> tuple[dict[str, str], dict[str, str]]:
    properties_vars: dict[str, str] = {}
    manage_deps: dict[str, str] = {}
    for pom_file_location in resolver.files_by_glob("**/pom.xml", "pom.xml"):
        content = resolver.file_contents_by_location(pom_file_location)
        if (
            content
            and (pom_project := get_pom_project(content))
            and _is_parent_pom(pom_project, parent_info)
            and _is_module_parent(pom_project, current_pom_file_path)
        ):
            properties_vars = _get_properties(pom_project)
            manage_deps = _get_deps_management(pom_project)
            break

    manage_deps.update(_get_deps_management(current_pom_project))
    return properties_vars, manage_deps


def is_version_correct(version: str | None) -> bool:
    return version is not None and not version.startswith("${")


def new_package_from_pom_xml(
    project: Tag,
    dependency: Tag,
    location: Location,
    parent_info: dict[str, str] | None,
    parent_version_properties: dict[str, str] | None,
) -> Package | None:
    name = _get_text(dependency, "artifactid")

    if not name:
        return None

    group_id = _get_text(dependency, "groupid")
    full_name = group_id + ":" + name if group_id else name

    java_archive = JavaArchive(
        pom_properties=JavaPomProperties(
            group_id=group_id,
            artifact_id=_get_text(dependency, "artifactid"),
            version=_get_text(dependency, "version") if dependency.version else None,
        ),
    )

    version = dependency.version.get_text() if dependency.version else None
    if (
        version
        and version.startswith("${")
        and (
            parent_version_node := project.find_next(
                extract_bracketed_text(version),
            )
        )
    ):
        version_text = parent_version_node.get_text()
        if version_text and not version_text.startswith("${"):
            version = version_text

    if version and version.startswith("${") and parent_info and parent_version_properties:
        version = parent_version_properties.get(extract_bracketed_text(version), None)

    if (
        not version
        and parent_info
        and java_archive.pom_properties
        and java_archive.pom_properties.group_id
        and java_archive.pom_properties.artifact_id
    ):
        version = recursively_find_versions_from_parent_pom(
            group_id=java_archive.pom_properties.group_id,
            artifact_id=java_archive.pom_properties.artifact_id,
            parent_group_id=parent_info["group"],
            parent_artifact_id=parent_info["artifact"],
            parent_version=parent_info["version"],
        )

    if not is_version_correct(version):
        return None

    if location.coordinates:
        location.coordinates.line = dependency.sourceline
        location.dependency_type = DependencyType.DIRECT

    try:
        return Package(
            name=full_name,
            version=str(version),
            licenses=[],
            locations=[location],
            language=Language.JAVA,
            type=PackageType.JavaPkg,
            metadata=java_archive,
            p_url=package_url(name, str(version), java_archive),
        )
    except ValidationError as ex:
        LOGGER.warning(
            "Malformed package. Required fields are missing or data types are incorrect.",
            extra={
                "extra": {
                    "exception": format_exception(str(ex)),
                    "location": location.path(),
                },
            },
        )
        return None


def parse_pom_xml(
    resolver: Resolver | None,
    _env: Environment | None,
    reader: LocationReadCloser,
) -> tuple[list[Package], list[Relationship]]:
    root = None
    with suppress(UnicodeError):
        try:
            root = BeautifulSoup(reader.read_closer, features="html.parser")
        except AssertionError:
            return [], []

    if not root:
        return [], []

    pkgs = []
    if (
        (project := root.project)
        and str(project.get("xmlns")) == "http://maven.apache.org/POM/4.0.0"
        and (dependencies := project.find("dependencies", recursive=False))
        and isinstance(dependencies, Tag)
    ):
        parent_info = _get_parent_info(project)
        parent_version_properties = None
        if resolver and parent_info:
            parent_version_properties, _ = _evaluate_pom_files_in_project(
                resolver,
                parent_info,
                str(reader.location.access_path),
                project,
            )

        for dependency in dependencies.find_all("dependency"):
            pkg = new_package_from_pom_xml(
                project,
                dependency,
                deepcopy(reader.location),
                parent_info,
                parent_version_properties,
            )
            if pkg:
                pkgs.append(pkg)
    return pkgs, []


def decode_pom_xml(content: str) -> Tag:
    return BeautifulSoup(content, features="html.parser")


def pom_parent(parent: Tag | None) -> JavaPomParent | None:
    if not parent:
        return None

    group_id = _get_text(parent, "groupId")
    artifact_id = _get_text(parent, "artifactId")
    version = _get_text(parent, "version")
    if not group_id or not artifact_id or not version:
        return None

    result = JavaPomParent(
        group_id=group_id,
        artifact_id=artifact_id,
        version=version,
    )

    if not result.group_id and not result.artifact_id and not result.version:
        return None

    return result


def parse_pom_xml_project(
    path: str,
    reader: str,
    _location: Location,
) -> ParsedPomProject | None:
    project = BeautifulSoup(reader, features="xml").project
    if not project:
        return None
    return new_pom_project(path, project, _location)


def _find_direct_child(parent: Tag, tag: str) -> Tag | None:
    return next(
        (child for child in parent.find_all(tag, recursive=False) if child.parent == parent),
        None,
    )


def new_pom_project(
    path: str,
    project: Tag,
    _location: Location,
) -> ParsedPomProject:
    artifact_id = _safe_string(_find_direct_child(project, "artifactId"))
    name = _safe_string(_find_direct_child(project, "name"))
    project_url = _safe_string(_find_direct_child(project, "url"))

    licenses: list[str] = []
    if project.licenses:
        for license_ in project.licenses.find_all("license"):
            license_name: str | None = None
            license_url: str | None = None
            if name_node := license_.find_next("name"):
                license_name = name_node.get_text()
            elif url_node := license_.find_next("url"):
                license_url = url_node.get_text()

            if not license_name and not license_url:
                continue
            if license_name:
                licenses.append(license_name)
            elif license_url:
                licenses.append(license_url)

    return ParsedPomProject(
        java_pom_project=JavaPomProject(
            path=path,
            parent=pom_parent(_find_direct_child(project, "parent")),
            group_id=_safe_string(_find_direct_child(project, "groupId")),
            artifact_id=artifact_id,
            version=_safe_string(_find_direct_child(project, "version")),
            name=name,
            description=_safe_string(
                _find_direct_child(project, "description"),
            ),
            url=project_url,
        ),
        licenses=licenses,
    )


def _safe_string(value: Tag | None) -> str:
    if not value:
        return ""
    return value.get_text()
